# !/usr/bin/env python
# -*- coding: UTF-8 -*-
""" Generic Facade to Find Data in 1..* Ontology Models """


from functools import lru_cache
from typing import Callable
from typing import Optional
from typing import List
from collections import defaultdict

from baseblock import BaseObject, Enforcer

from owl_finder.multiquery.dmo import (ModelResultMerge, NerPalleteLookup,
                                       ViewGeneratorLookup)
from owl_finder.multiquery.svc import (FindNER, FindSynonyms, FindTypes,
                                       LoadSynonyms, QueryNerDepth,
                                       QueryNerLabel, QueryNerTaxo)
from owl_finder.singlequery.bp import AskOwlAPI
from owl_finder.singlequery.dto import QueryResultType


class FindOntologyData(BaseObject):
    """ Generic Facade to Find Data in 1..* Ontology Models """

    def __init__(self,
                 ontologies: list,
                 absolute_path: str):
        """ Change Log

        Created:
            26-May-2022
            craigtrim@gmail.com
            *   https://github.com/grafflr/deepnlu/issues/2
        Updated:
            24-Nov-2022
            craigtrim@gmail.com
            *   correct casing in labels-rev
                https://github.com/craigtrim/owl-finder/issues/1#issuecomment-1326750848
            *   corrected defect in children-and-self
                https://github.com/craigtrim/owl-finder/issues/1#issuecomment-1326754158

        Args:
            ontologies (list): one-or-more Ontology models to use in processing
            absolute_path (str): an absolute path that applies to all the OWL models
        """
        BaseObject.__init__(self, __name__)
        if self.isEnabledForDebug:
            Enforcer.is_list(ontologies)

        self._ontologies = ontologies
        self._absolute_path = absolute_path

        self._d_ontologies = self._load(
            ontologies=ontologies,
            absolute_path=absolute_path)

        self._merge = ModelResultMerge().process

        self._query_ner_label = QueryNerLabel(self._d_ontologies).process
        self._query_ner_depth = QueryNerDepth(self._d_ontologies).process
        self._query_ner_taxo = QueryNerTaxo(self._d_ontologies).process

        self._load_synonyms = LoadSynonyms(
            d_ontologies=self._d_ontologies,
            model_result_merge=self._merge)

        self._find_synonyms = FindSynonyms(
            d_synonyms_fwd=self.synonyms(),
            d_synonyms_rev=self.synonyms_rev())

        self._find_types = FindTypes(
            d_types_fwd=self.types(),
            d_types_rev=self.types_rev())

    def ontologies(self) -> list:
        return self._ontologies

    def absolute_path(self) -> str:
        return self._absolute_path

    def _load(self,
              ontologies: list,
              absolute_path: str) -> dict:
        d = {}
        for ontology_name in ontologies:
            d[ontology_name] = AskOwlAPI(
                ontology_name=ontology_name,
                absolute_path=absolute_path)
        return d

    @staticmethod
    def _to_entity_name(input_text: str) -> str:
        input_text = input_text.lower().strip()
        if ' ' in input_text:
            input_text = input_text.replace(' ', '_')
        return input_text

    @staticmethod
    def _sort_list(values: list) -> list:
        return sorted(set(values), reverse=False)

    @staticmethod
    def _type_check(input_text: str) -> None:
        """ Perform Consistent Type Checking

        Args:
            input_text (str): any input text

        Raises:
            TypeError: a type error if the input is empty
        """
        if not input_text or not len(input_text):
            raise TypeError

    def by_predicate(self,
                     predicate_name: str,
                     to_lowercase: bool = True) -> dict:
        """ Create a Subject-to-Objects dictionary by Predicate Name

        Sample Ontology:
            ###  http://graffl.ai/skills#Artifact
            :Artifact rdf:type owl:Class ;
                    rdfs:label "Artifact" ;
                    owl:backwardCompatibleWith "ARTIFACT"^^:grafflNER .

        Sample Output:
            {
                'Artifact': ['ARTIFACT']
            }

        Args:
            predicate_name (str): the name of the predicate
            to_lowercase (bool, optional): enforce case sensitive search. Defaults to True.

        Returns:
            dict: triples (keyed by subject)
        """
        return self._by_predicate(predicate_name, to_lowercase)

    @lru_cache
    def _by_predicate(self,
                      predicate_name: str,
                      to_lowercase: bool = True) -> dict:
        results = []
        for ontology_name in self._d_ontologies:
            results.append(
                self._d_ontologies[ontology_name].by_predicate(
                    predicate=predicate_name,
                    to_lowercase=to_lowercase))

        if not results:
            return None
        elif len(results) == 1:
            return results[0]
        return self._merge(results, QueryResultType.DICT_OF_STR2LIST)

    def by_predicate_rev(self,
                         predicate_name: str,
                         to_lowercase: bool = True) -> dict:
        """ Create a Object-to-Subjects dictionary by Predicate Name

        Sample Ontology:
            ###  http://graffl.ai/skills#Artifact
            :Artifact rdf:type owl:Class ;
                    rdfs:label "Artifact" ;
                    owl:backwardCompatibleWith "ARTIFACT"^^:grafflNER .

        Sample Output:
            {
                'ARTIFACT': ['Artifact']
            }

        Args:
            predicate_name (str): the name of the predicate
            to_lowercase (bool, optional): enforce case sensitive search. Defaults to True.

        Returns:
            dict: triples (keyed by object)
        """
        return self._by_predicate_rev(predicate_name, to_lowercase)

    @lru_cache
    def _by_predicate_rev(self,
                          predicate_name: str,
                          to_lowercase: bool = True) -> dict:
        results = []
        for ontology_name in self._d_ontologies:
            results.append(
                self._d_ontologies[ontology_name].by_predicate(
                    reverse=True,
                    predicate=predicate_name,
                    to_lowercase=to_lowercase))

        if not results:
            return None
        elif len(results) == 1:
            return results[0]
        return self._merge(results, QueryResultType.DICT_OF_STR2LIST)

    def transitive(self,
                   input_text: str,
                   query: Callable) -> list:
        """ Invoke any function in a transitive (recursive) manner

        Usage:
            self.transitive('<entity>', self.implies_by_entity)

        Args:
            input_text (str): any input text or entity
            query (Callable): the function to use recursively

        Returns:
            list: the results (if any)
        """
        results = []

        def update(query_results: list or None):
            if query_results:
                [results.append(x) for x in query_results]

        update(query(input_text))
        for parent in self.parents(input_text):
            update(self.transitive(parent, query))

        return results

    @lru_cache
    def comments(self) -> dict:
        results = []
        for ontology_name in self._d_ontologies:
            results.append(self._d_ontologies[ontology_name].comments())

        if not results:
            return None
        elif len(results) == 1:
            return results[0]
        return self._merge(results, QueryResultType.DICT_OF_STR2LIST)

    def effects(self) -> dict:
        """ Find Effects Relationships

        Sample:
            Given   ?a effects ?b
            Return  ?a : [?b]

        Returns:
            list or None: a list of zero-or-more entities that the incoming entity effects
        """
        return self._by_predicate('effects')

    @lru_cache
    def effects_rev(self) -> dict:
        """ Find Effected By relationships

        Sample:
            Given   ?a effected by ?b
            Return  ?b : [?a]

        Returns:
            dict: reverse of effected by function
        """
        return self._by_predicate_rev('effects')

    @lru_cache
    def requires(self) -> dict:
        """ Find Requires Relationships

        Sample:
            Given   ?a requires ?b
            Return  ?a : [?b]

        Returns:
            list or None: a list of zero-or-more entities that the incoming entity requires
        """
        return self._by_predicate('requires')

    @lru_cache
    def required_by(self) -> dict:
        """ Find Required By relationships

        Sample:
            Given   ?a requires ?b
            Return  ?b : [?a]

        Returns:
            dict: reverse of requires function
        """
        return self._by_predicate_rev('requires')

    def requires_by_entity(self,
                           input_text: str) -> list or None:
        """ Retrict Requires Relationship to a Single Entity

        Args:
            input_text (str): the entity to search for

        Returns:
            list or None: a list of zero-or-more the incoming entity requires
        """
        input_text = self._to_entity_name(input_text)
        if self.requires() and input_text in self.requires():
            return self.requires()[input_text]

    def required_by_entity(self,
                           input_text: str) -> list or None:
        """ Retrict Required-By Relationship to a Single Entity

        Args:
            input_text (str): the entity to search for

        Returns:
            list or None: a list of zero-or-more entities that the incoming entity is required by
        """
        input_text = self._to_entity_name(input_text)
        if self.required_by() and input_text in self.required_by():
            return self.required_by()[input_text]

    @lru_cache
    def similar(self) -> dict:
        """ Find Similar Relationships

        Sample:
            Given       ?a similar ?b
            Return      ?a : [?b]

        Returns:
            list or None: a list of zero-or-more entities that the incoming entity is similar to
        """
        return self._by_predicate('similarTo')

    @lru_cache
    def similar_rev(self) -> dict:
        """ Find Similar Inverse Relationships

        Sample:
            Given       ?a similar ?b
            Return      ?b : [?a]

        Returns:
            list or None: a list of zero-or-more entities that the incoming entity is similar to
       """
        return self._by_predicate_rev('similarTo')

    def similar_by_entity(self,
                          input_text: str) -> dict:
        """ Find Similarity Relationships on a per-Entity basis
        This relationship is bi-directional

        Sample:
            Given       ?a similar ?b
                        ?x similar ?a
            Return      ?a : [?b, ?x]

        Rationale:
            if      ?x is similar to ?a
            then    ?a is similar to ?x

        Args:
            input_text (str): the entity to search for

        Returns:
            list or None: a list of zero-or-more entities that the incoming entity is similar to
        """
        results = []
        input_text = self._to_entity_name(input_text)

        if self.similar() and input_text in self.similar():
            [results.append(x) for x in self.similar()[input_text]]

        if self.similar_rev() and input_text in self.similar_rev():
            [results.append(x) for x in self.similar_rev()[input_text]]

        return results

    @lru_cache
    def implies(self) -> dict:
        """ Find Implied Relationships

        Sample:
            Given       ?a implies ?b
            Return      ?a : [?b]

        Returns:
            list or None: a list of zero-or-more entities that the incoming entity implies
        """
        return self._by_predicate('implies')

    @lru_cache
    def implies_by_entity(self,
                          input_text: str) -> dict:
        """ Retrict Implies Relationship to a Single Entity

        Args:
            input_text (str): the entity to search for

        Returns:
            list or None: a list of zero-or-more entities that the incoming entity implies
        """
        input_text = self._to_entity_name(input_text)
        if self.implies() and input_text in self.implies():
            return self.implies()[input_text]

    @lru_cache
    def implied_by(self) -> dict:
        """ Find Implied By relationships

        Sample:
            Given   ?a requires ?b
            Return  ?b : [?a]

        Returns:
            list or None: a list of zero-or-more entities that are implied by this entity
        """
        return self._by_predicate_rev('implies')

    @lru_cache
    def implied_by_entity(self,
                          input_text: str) -> dict:
        """ Retrict Implied By Relationship to a Single Entity

        Args:
            input_text (str): the entity to search for

        Returns:
            list or None: a list of zero-or-more entities that are implied by this entity
        """
        input_text = self._to_entity_name(input_text)
        if self.implied_by() and input_text in self.implied_by():
            return self.implied_by()[input_text]

    @lru_cache(maxsize=512, typed=False)
    def is_canon(self,
                 input_text: str) -> bool:
        """Check if Input Text is Canonical Entity

        Args:
            input_text (str): any input string

        Returns:
            bool: True if the input string is a Nursing Entity
        """
        return self._find_synonyms.is_canon(input_text)

    @lru_cache(maxsize=512, typed=False)
    def find_canon(self,
                   input_text: str) -> str or None:
        """Find the Canonical Representation of the Input String

        Args:
            input_text (str): any input string

        Returns:
            str or None: The Canonical Entity
        """
        return self._find_synonyms.find_canon(input_text)

    @lru_cache(maxsize=512, typed=False)
    def is_variant(self,
                   input_text: str) -> bool:
        """Check if Input Text is known variant for at least one Canonical Entry

        Args:
            input_text (str): any input string

        Returns:
            bool: True if the input string is a known synonym to a Nursing Entity
        """
        return self._find_synonyms.is_variant(input_text)

    @lru_cache(maxsize=512, typed=False)
    def find_variants(self,
                      input_text: str) -> list or None:
        """Find the Synonyms for a known Entity

        Args:
            input_text (str): any input string

        Returns:
            list or None: a list of synonyms for the input entity
        """
        return self._find_synonyms.find_variants(input_text)

    @lru_cache(maxsize=512, typed=False)
    def find_ner(self,
                 input_text: str) -> str or None:

        svc = FindNER(
            d_ner_depth=self.ner_depth(),
            d_ner_taxo=self.ner_taxonomy(),
            d_graffl_ner=self.graffl_ner(),
            d_spacy_ner=self.spacy_ner())

        return svc.find_ner(input_text)

    @lru_cache
    def graffl_ner(self) -> dict:
        return self._query_ner_label('grafflNER')

    @lru_cache
    def graffl_ner_rev(self) -> dict:
        return self._query_ner_label('grafflNER', reverse=True)

    @lru_cache
    def spacy_ner(self) -> dict:
        return self._query_ner_label('spacyNER')

    @lru_cache
    def spacy_ner_rev(self) -> dict:
        return self._query_ner_label('spacyNER', reverse=True)

    @lru_cache
    def ner_depth(self) -> dict:
        return self._query_ner_depth(reverse=False)

    @lru_cache
    def ner_depth_rev(self) -> dict:
        return self._query_ner_depth(reverse=True)

    @lru_cache
    def infer_by_requires(self) -> dict:
        raise NotImplementedError

    @lru_cache
    def labels(self) -> Optional[dict]:
        """Find all the Labels keyed by Entity Name

        Returns:
            Optional[dict]: a list of entity names with label string values
        """
        # the challenge with using 'to_lowercase=False' is that both the key and value retain their original case
        d_labels = self._by_predicate('rdfs:label', to_lowercase=False)
        if not d_labels or not len(d_labels):
            return None

        d = {}
        for k in d_labels:
            d[k.lower()] = d_labels[k]

        return d

    @lru_cache
    def labels_rev(self) -> Optional[dict]:
        """Find all the Entity Names keyed by Label

        Notes:
            I need to maintain the integrity of label casing (preserved) vs
            entity-casing (enforced lower-case).  As a result, I finder it easier
            to perform a custom reversal in this function rather than use 'by_predicate_rev'

        Returns:
            Optional[dict]: a list of labels with entity name values
        """
        d_labels = self.labels()
        if not d_labels or not len(d_labels):
            return None

        d_rev = defaultdict(list)
        for entity in d_labels:
            for label in d_labels[entity]:
                d_rev[label].append(entity)

        return dict(d_rev)

    def label_by_entity(self,
                        input_text: str) -> str or None:
        """Find the Label for a known Entity

        Args:
            input_text (str): any input string

        Returns:
            str or None: a label for the entity (if found)
        """
        input_text = self._to_entity_name(input_text)
        if input_text not in self.labels():
            pass

        if input_text in self.labels():
            results = self.labels()[input_text]
            if results and len(results):
                return results[0]

    @lru_cache
    def ner_taxonomy(self) -> dict:
        return self._query_ner_taxo(reverse=False)

    @lru_cache
    def ner_taxonomy_rev(self) -> dict:
        return self._query_ner_taxo(reverse=True)

    @lru_cache
    def ner_pallete_lookup(self,
                           input_text: str) -> dict or None:
        d_result = self.ner_taxonomy_rev()

        if not d_result or not len(d_result):
            return None

        return NerPalleteLookup(d_result).lookup(input_text)

    @lru_cache
    def ner_pallete_colors(self) -> dict or None:
        d_result = self.ner_taxonomy_rev()

        if not d_result or not len(d_result):
            return None

        return NerPalleteLookup(d_result).colors()

    @lru_cache
    def spans(self) -> dict:
        results = []
        for ontology_name in self._d_ontologies:
            results.append(self._d_ontologies[ontology_name].spans())

        if not results:
            return None
        elif len(results) == 1:
            return results[0]

        return self._merge(results, QueryResultType.DICT_OF_STR2DICT)

    @lru_cache
    def span_keys(self) -> Optional[list]:
        """ Return Span Keys sorted by Length

        Returns:
            list: list of span keys sorted by length
        """
        spans = self.spans()
        if spans and len(spans):
            return sorted(self.spans().keys(), key=len)

    def synonyms(self) -> dict:
        """ Return Synonyms keyed by Entity Name

        Returns:
            dict: dictionary of entities associated to a list of zero-or-more synonyms
        """
        return self._load_synonyms.synonyms()

    def synonyms_rev(self) -> dict:
        """ Return Entities keyed by Synonyms

        Returns:
            dict: dictionary of synonyms associated to a list of one-or-more entities
        """
        return self._load_synonyms.synonyms_rev()

    @lru_cache
    def trie(self) -> dict:
        results = []
        for ontology_name in self._d_ontologies:
            results.append(self._d_ontologies[ontology_name].trie())

        if not results:
            return None
        elif len(results) == 1:
            return results[0]
        return self._merge(results, QueryResultType.DICT_OF_STR2LIST)

    @lru_cache
    def types(self) -> dict:
        return self._by_predicate('rdfs:subClassOf')

    @lru_cache
    def types_rev(self) -> dict:
        return self._by_predicate_rev('rdfs:subClassOf')

    @lru_cache
    def uses(self) -> dict:
        return self._by_predicate('uses')

    @lru_cache
    def uses_rev(self) -> dict:
        return self._by_predicate_rev('uses')

    @lru_cache
    def has_parent(self,
                   input_text: str,
                   parent: str) -> bool:
        """ Simple Truth check
            Does the incoming input entity have the incoming parent as a parent?

        Args:
            input_text (str): a candidate concept

        Returns:
            bool: True if the described relationship exists
        """
        return self._find_types.has_parent(
            parent=parent,
            input_text=input_text)

    @lru_cache
    def has_ancestor(self,
                     input_text: str,
                     parent: str) -> bool:
        """ Simple Truth check
            Does the incoming input entity have the incoming parent as an ancestor?

        Args:
            input_text (str): a candidate concept

        Returns:
            bool: True if the described relationship exists
        """
        return self._find_types.has_ancestor(
            parent=parent,
            input_text=input_text)

    @lru_cache
    def entity_exists(self,
                      input_text: str) -> bool:
        """ Simple Truth check
            Does this value exist anywhere in the Ontology?

        Args:
            input_text (str): a candidate concept

        Returns:
            bool: True if the concept exists in the Ontology
        """
        return self._find_types.exists(input_text)

    @lru_cache
    def children(self,
                 input_text: str) -> List[str]:
        """ Return the Children for an Entity

        Args:
            input_text (str): the input entity

        Raises:
            TypeError: Invalid Input

        Returns:
            List[str]: the results
        """
        self._type_check(input_text)
        return self._sort_list(self._find_types.children(input_text))

    @lru_cache
    def children_and_self(self,
                          input_text: str) -> List[str]:
        """ Return the Children for an Entity,
            and the Entity itself, in a list of results

        Args:
            input_text (str): the input entity

        Raises:
            TypeError: Invalid Input

        Returns:
            List[str]: the results
        """
        self._type_check(input_text)
        if not self.entity_exists(input_text):
            return []

        input_text = input_text.lower().strip()
        results = self._find_types.children(input_text)
        if not results:
            results = []

        results.append(input_text)
        return self._sort_list(results)

    @lru_cache
    def descendants(self,
                    input_text: str) -> list:
        """ Return the Descendants for an Entity

        Args:
            input_text (str): the input entity

        Raises:
            TypeError: Invalid Input

        Returns:
            list: the list of results (if any)
        """
        self._type_check(input_text)
        return self._find_types.descendants(input_text)

    @lru_cache
    def descendants_and_self(self,
                             input_text: str) -> list:
        """ Return the Descendants for an Entity,
            and the Entity itself, in a list of results

        Args:
            input_text (str): the input entity

        Raises:
            TypeError: Invalid Input

        Returns:
            list: the list of results (if any)
        """
        self._type_check(input_text)
        if not self.entity_exists(input_text):
            return []

        input_text = input_text.lower().strip()
        results = self._find_types.descendants(input_text)
        if not results:
            results = []

        results.append(input_text)
        return self._sort_list(results)

    @lru_cache
    def parents(self,
                input_text: str) -> list:
        """ Return the Parents for an Entity

        Args:
            input_text (str): the input entity

        Raises:
            TypeError: Invalid Input

        Returns:
            list: the list of results (if any)
        """
        self._type_check(input_text)
        return self._find_types.parents(input_text)

    @lru_cache
    def parents_and_self(self,
                         input_text: str) -> list:
        """ Return the Parents for an Entity,
            and the Entity itself, in a list of results

        Args:
            input_text (str): the input entity

        Raises:
            TypeError: Invalid Input

        Returns:
            list: the list of results (if any)
        """
        self._type_check(input_text)
        if not self.entity_exists(input_text):
            return []

        input_text = input_text.lower().strip()
        results = self._find_types.parents(input_text)
        if not results:
            results = []

        results.append(input_text)
        return self._sort_list(results)

    @lru_cache
    def ancestors(self,
                  input_text: str) -> list:
        """ Return the Ancestors for an Entity

        Args:
            input_text (str): the input entity

        Raises:
            TypeError: Invalid Input

        Returns:
            list: the list of results (if any)
        """
        self._type_check(input_text)
        return self._sort_list(self._find_types.ancestors(input_text))

    @lru_cache
    def ancestors_and_self(self,
                           input_text: str) -> list:
        """ Return the Ancestors for an Entity,
            and the Entity itself, in a list of results

        Args:
            input_text (str): the input entity

        Raises:
            TypeError: Invalid Input

        Returns:
            list: the list of results (if any)
        """
        self._type_check(input_text)
        if not self.entity_exists(input_text):
            return []

        input_text = input_text.lower().strip()
        results = self._find_types.ancestors(input_text)
        if not results:
            results = []

        results.append(input_text)
        return self._sort_list(results)

    @lru_cache
    def lookup(self) -> dict or None:
        """ Generate n-Gram Spans suitable for Synonym Matching

        Reference:
            https://github.com/grafflr/ask-owl/issues/4
            https://github.com/grafflr/deepnlu/issues/21#issuecomment-1141524102

        Returns:
            dict: dictionary of values keyed by n-gram size
        """
        d_synonyms_fwd = self.synonyms()

        if not d_synonyms_fwd or not len(d_synonyms_fwd):
            return None

        return ViewGeneratorLookup().process(d_synonyms_fwd)

    def has_data(self) -> bool:
        """ Check if the underlying Ontologies have data

        Returns:
            bool: _description_
        """
        return self.synonyms() and len(self.synonyms())



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
""" View Generator: Entity Spans for Long-Range Synonym Swapping """


from pprint import pprint
from typing import List

from collections import defaultdict

from baseblock import EnvIO
from baseblock import BaseObject


class GeneratePlusSpans(BaseObject):
    """ View Generator: Entity Spans for Long-Range Synonym Swapping """

    def __init__(self):
        """ Change Log

        Created:
            25-Nov-2022
            craigtrim@gmail.com
            *   https://github.com/craigtrim/owl-finder/issues/4
        Updated:
            28-Nov-2022
            craigtrim@gmail.com
            *   use hash to ensure uniqueness in spans
                https://github.com/craigtrim/owl-finder/issues/4#issuecomment-1329990189
        """
        BaseObject.__init__(self, __name__)

    def _to_dict(self,
                 canon: str,
                 tokens: list) -> dict:

        tokens = sorted(set(tokens), reverse=False)
        if not len(tokens):
            return None

        # TODO: all these values have to be default only when not otherwise specified
        distance = EnvIO.int_or_default('SPAN_DISTANCE', 3)
        return {
            'content': tokens,
            'distance': distance,
            'forward': True,
            'reverse': True,
            'canon': canon
        }

    def process(self,
                d_results: dict) -> dict:

        hashes = set()
        d_spans = defaultdict(list)

        for entity in d_results:

            synonyms = d_results[entity]
            synonyms = [x for x in synonyms if '+' in x]

            for synonym in synonyms:

                tokens = synonym.split('+')
                d_result = self._to_dict(entity, tokens[1:])

                # 20221128; Ensure Uniqueness
                # https://github.com/craigtrim/owl-finder/issues/4#issuecomment-1329990189
                hash_d_result = hash(''.join([
                    str(x) for x in d_result.values()
                ]))

                if hash_d_result not in hashes:
                    d_spans[tokens[0]].append(d_result)
                    hashes.add(hash_d_result)

        return dict(d_spans)



from .external_synonym_loader import ExternalSynonymLoader
from .model_result_merge import ModelResultMerge
from .ner_pallete_lookup import NerPalleteLookup
from .view_generator_lookup import ViewGeneratorLookup
from .view_generator_nerdepth import ViewGeneratorNerDepth
from .view_generator_nerlabel import ViewGeneratorNerLabel
from .view_generator_nertaxo import ViewGeneratorNerTaxo



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
""" Query the in-memory Ontology Model """


from collections import defaultdict

from baseblock import BaseObject, Stopwatch
from rdflib import Graph

from owl_finder.singlequery.dmo import OwlQueryExtract
from owl_finder.singlequery.dto import QueryResultType


class QueryOntologyModel(BaseObject):
    """ Query the in-memory Ontology Model """

    def __init__(self,
                 graph: Graph):
        """ Change Log

        Created:
            25-May-2022
            craigtrim@gmail.com
            *   https://github.com/craigtrim/askowl/issues/1
        Updated:
            27-May-2022
            craigtrim@gmail.com
            *   expose ability to lowercase output
                https://github.com/craigtrim/askowl/issues/4

        Args:
            graph (Graph): the instantiated RDF graph
        """
        BaseObject.__init__(self, __name__)
        self._execute_query = OwlQueryExtract(graph).process

    @staticmethod
    def _reverse_order(d_results: dict,
                       result_type: QueryResultType) -> dict:

        if result_type == QueryResultType.DICT_OF_STR2LIST:
            d_rev = defaultdict(list)
            for k in d_results:
                for v in d_results[k]:
                    d_rev[v].append(k)
            return dict(d_rev)

        raise NotImplementedError

    def process(self,
                sparql: str,
                result_type: QueryResultType,
                reverse: bool = False,
                to_lowercase: bool = True) -> dict or list:
        """ Execute a SPARQL query on the RDF Graph

        Args:
            sparql (str): the SPARQL query to execute
            result_type (QueryResultType): the type of transformation to perform on the result set
            reverse (bool, optional): reverses the subject/object order. Defaults to False.
                if "?x implies ?y" and reverse=False
                    the results will be { ?x: [?y-1, ?y-2, ..., ?y-N]}
                if "?x implies ?y" and reverse=True
                    the results will be { ?y-1: [x], ?y-2: [x], ?y-N: [x]}
            to_lowercase (bool, optional): Ensures all output is lower-cased. Defaults to True.

        Returns:
            dict or list: the result set
        """

        sw = Stopwatch()

        d_results = self._execute_query(
            query=sparql,
            to_lowercase=to_lowercase,
            result_type=result_type)

        if not d_results or not len(d_results):
            if self.isEnabledForDebug:
                self.logger.info('\n'.join([
                    'Ontology Model Service Completed',
                    f'\tTotal Time: {str(sw)}',
                    '\t*** No Results Found ***',
                    f'\tSPARQL Query: {sparql}']))
            return None

        if reverse:
            d_results = self._reverse_order(
                d_results=d_results,
                result_type=result_type)

        if self.isEnabledForInfo:
            self.logger.info('\n'.join([
                'Ontology Model Service Completed',
                f'\tTotal Time: {str(sw)}',
                f'\tTotal Results: {len(d_results)}',
                f'\tSPARQL Query: {sparql}']))

        return d_results



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
""" Create Ontology Spans using N-Gram Input """


from collections import defaultdict
from pprint import pprint

from baseblock import BaseObject


class OwlSpanGenerate(BaseObject):
    """ Create Ontology Spans using N-Gram Input """

    __punkt = [
        '!',
        '?',
        '.',
    ]

    def __init__(self):
        """ Change Log

        Created:
            26-May-2022
            craigtrim@gmail.com
            *   https://github.com/grafflr/ask-owl/issues/4
        Updated:
            1-Jun-2022
            craigtrim@gmail.com
            *   add dedupe function
                https://github.com/grafflr/deepnlu/issues/28#issuecomment-1144270515

        """
        BaseObject.__init__(self, __name__)

    def _to_gramsize_dict(self,
                          d_results: dict) -> dict:
        d = defaultdict(list)
        for k in d_results:

            name = k.replace('_', ' ')
            gram_size = len(name.split(' '))
            d[gram_size].append(name)

            for v in d_results[k]:
                name = v.replace('_', ' ')
                gram_size = len(name.split(' '))
                d[gram_size].append(name)

                for punkt in self.__punkt:  # GRAFFL-155
                    if v.endswith(punkt):
                        d[gram_size].append(v[:len(v) - len(punkt)])

        return d

    @staticmethod
    def _lower(d_results: dict) -> dict:
        d = {}
        for k in d_results:
            d[k.lower()] = [v.lower().strip() for v in d_results[k]]

        return d

    @staticmethod
    def _dedupe(d_results: dict) -> dict:
        return {k: sorted(set(d_results[k])) for k in d_results}

    def process(self,
                d_results: dict) -> dict:

        d_results = self._lower(d_results)
        d_gramsize = self._to_gramsize_dict(d_results)

        # Rationale: https://github.com/grafflr/deepnlu/issues/28#issuecomment-1144270515
        d_gramsize = self._dedupe(d_gramsize)

        # enforce keys
        for i in range(1, 7):
            if i not in d_gramsize:
                d_gramsize[i] = []

        return d_gramsize



# !/usr/bin/env python
# -*- coding: UTF-8 -*-
""" Generic Facade to interact with Entity Taxonomies """


from baseblock import BaseObject


class FindTypes(BaseObject):
    """ Generic Facade to interact with Entity Taxonomies """

    def __init__(self,
                 d_types_fwd: dict,
                 d_types_rev: dict):
        """ Change Log

        Created:
            29-Oct-2021
            craigtrim@gmail.com
            *   https://github.com/grafflr/graffl-core/issues/97
        Updated:
            24-Oct-2022
            craigtrim@gmail.com
            *   Update Finder Methods to Replace Spaces with Underscores
                when no other results are found
                https://github.com/grafflr/graffl-core/issues/129
        Updated:
            25-Jan-2022
            craigtrim@gmail.com
            *   pass ontology-name as optional param
                https://github.com/grafflr/graffl-core/issues/135
        Updated:
            1-Feb-2022
            craigtrim@gmail.com
            *   make ontology param consistent; permit multiple values
                https://github.com/grafflr/graffl-core/issues/135#issuecomment-1027468040
            *   a finder initialization is a contract
                https://github.com/grafflr/graffl-core/issues/135#issuecomment-1027474785
        Updated:
            26-May-2022
            craigtrim@gmail.com
            *   treat 'ontologies' param as a list
                https://github.com/grafflr/deepnlu/issues/7
        Updated:
            27-May-2022
            craigtrim@gmail.com
            *   migrated to 'owlblock' in pursuit of
                https://github.com/grafflr/deepnlu/issues/13

        Args:
            ontologies (list): one-or-more Ontology models to use in processing
        """
        BaseObject.__init__(self, __name__)
        self._d_types_fwd = d_types_fwd
        self._d_types_rev = d_types_rev

    def _has_types_fwd(self) -> bool:
        """ No Types Exist
        This result is unusual outside of unit-testing; but could happen

        It is possible to have a "flat" Taxonomy within the OWL model
            where no parent/child (rdfs:subClassOf) relationships exist

        Reference:
            https://github.com/grafflr/deepnlu/issues/20

        Returns:
            bool: True if types exist
        """
        return self._d_types_fwd and len(self._d_types_fwd)

    def _has_types_rev(self) -> bool:
        """ No Types Exist
        This result is unusual outside of unit-testing; but could happen

        It is possible to have a "flat" Taxonomy within the OWL model
            where no parent/child (rdfs:subClassOf) relationships exist

        Reference:
            https://github.com/grafflr/deepnlu/issues/20

        Returns:
            bool: True if types exist
        """
        return self._d_types_rev and len(self._d_types_rev)

    def has_parent(self,
                   input_text: str,
                   parent: str) -> bool:
        input_text = input_text.lower().strip()
        return parent in self.parents(input_text)

    def has_ancestor(self,
                     input_text: str,
                     parent: str) -> bool:
        input_text = input_text.lower().strip()
        return parent in self.ancestors(input_text)

    def exists(self,
               input_text: str) -> bool:
        """ Simple Truth check
            Does this value exist anywhere in the Ontology?

        Args:
            input_text (str): a candidate concept

        Returns:
            bool: True if the concept exists in the Ontology
        """

        if not self._has_types_fwd() and not self._has_types_rev():
            return False

        input_text = input_text.lower().strip()
        if ' ' in input_text:
            input_text = input_text.replace(' ', '_')

        if self._has_types_fwd() and input_text in self._d_types_fwd:
            return True
        if self._has_types_rev() and input_text in self._d_types_rev:
            return True

        return bool(len(self.parents(input_text)))

    def children(self,
                 input_text: str) -> list:
        """ Return the Children for an Entity

        Args:
            input_text (str): the input entity

        Returns:
            list: the list of results (if any)
        """

        if not self._has_types_rev():
            return []

        input_text = input_text.lower().strip()

        if input_text in self._d_types_rev:
            return self._d_types_rev[input_text]

        if ' ' in input_text:
            return self.children(input_text.replace(' ', '_'))

        return []

    def children_and_self(self,
                          input_text: str) -> list:
        s = set()
        s.add(input_text)
        [s.add(x) for x in self.children(input_text)]
        return sorted(s)

    def descendants(self,
                    input_text: str) -> list:
        """ Return the Descendants for an Entity

        Args:
            input_text (str): the input entity

        Returns:
            list: the list of results (if any)
        """

        if not self._has_types_rev():
            return []

        results = []
        input_text = input_text.lower().strip()

        def _descendants(entity: str):
            if entity in self._d_types_rev:
                for child in self._d_types_rev[entity]:
                    results.append(child)
                    _descendants(child)

        _descendants(input_text)

        if not len(results) and ' ' in input_text:
            return self.descendants(input_text.replace(' ', '_'))

        return results

    def descendants_and_self(self,
                             input_text: str) -> list:
        s = set()
        s.add(input_text)
        [s.add(x) for x in self.descendants(input_text)]
        return sorted(s)

    def parents(self,
                input_text: str) -> list:
        """ Return the Parents for an Entity

        Args:
            input_text (str): the input entity

        Returns:
            list: the list of results (if any)
        """

        if not self._has_types_fwd():
            return []

        input_text = input_text.lower().strip()

        if input_text in self._d_types_fwd:
            return self._d_types_fwd[input_text]

        if ' ' in input_text:
            return self.parents(input_text.replace(' ', '_'))

        return []

    def parents_and_self(self,
                         input_text: str) -> list:
        s = set()
        s.add(input_text)
        [s.add(x) for x in self.parents(input_text)]
        return sorted(s)

    def ancestors(self,
                  input_text: str) -> list:
        """ Return the Ancestors for an Entity

        Args:
            input_text (str): the input entity

        Returns:
            list: the list of results (if any)
        """

        if not self._has_types_fwd():
            return []

        results = []
        input_text = input_text.lower().strip()

        def _ancestors(entity: str):
            if entity in self._d_types_fwd:
                for parent in self._d_types_fwd[entity]:
                    results.append(parent)
                    _ancestors(parent)

        _ancestors(input_text)

        if not len(results) and ' ' in input_text:
            return self.ancestors(input_text.replace(' ', '_'))

        return results

    def ancestors_and_self(self,
                           input_text: str) -> list:
        s = set()
        s.add(input_text)
        [s.add(x) for x in self.ancestors(input_text)]
        return sorted(s)



# !/usr/bin/env python
# -*- coding: UTF-8 -*-
""" Merge Multiple Model results into a single and cohesive structure """


from collections import defaultdict

from baseblock import BaseObject

from owl_finder.singlequery.dto import QueryResultType


class ModelResultMerge(BaseObject):
    """ Merge Multiple Model results into a single and cohesive structure """

    def __init__(self):
        """ Change Log

        Created:
            26-May-2022
            craigtrim@gmail.com
            *   https://github.com/grafflr/deepnlu/issues/2
        Updated:
            24-Nov-2022
            craigtrim@gmail.com
            *   sort the merged output
                https://github.com/craigtrim/owl-finder/issues/3
        """
        BaseObject.__init__(self, __name__)

    def _merge(self,
               results: list,
               result_type: QueryResultType) -> dict:

        results = [x for x in results if x is not None and len(x)]

        if result_type == QueryResultType.DICT_OF_STR2LIST:
            d_merge = defaultdict(list)
            for d_result in results:
                for k in d_result:
                    if k not in d_merge:
                        d_merge[k] = []

                    for value in d_result[k]:
                        if value not in d_merge[k]:
                            d_merge[k].append(value)

            return dict(d_merge)

        elif result_type == QueryResultType.DICT_OF_STR2DICT:
            d_merge = defaultdict(list)
            for d_result in results:
                for k in d_result:
                    for value in d_result[k]:
                        d_merge[k].append(value)

            return dict(d_merge)

        else:
            raise NotImplementedError

    def process(self,
                results: list,
                result_type: QueryResultType) -> list:

        d_model = self._merge(
            results=results,
            result_type=result_type)

        # Sort by Alpha
        # Reference https://github.com/craigtrim/owl-finder/issues/3
        d_sorted = {k: sorted(set(d_model[k]), reverse=False) for k in d_model}

        return d_sorted



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
""" View Generator: Generate NER Taxonomy """


from collections import defaultdict

from baseblock import BaseObject


class ViewGeneratorNerTaxo(BaseObject):
    """ View Generator: Generate NER Taxonomy """

    def __init__(self):
        """ Change Log

        Created:
            27-Oct-2021
            craig.@graffl.ai
            *   https://github.com/grafflr/graffl-core/issues/94
        Updated:
            27-May-2022
            craigtrim@gmail.com
            *   ported to 'deepnlu'
                https://github.com/grafflr/deepnlu/issues/12
        """
        BaseObject.__init__(self, __name__)

    @staticmethod
    def _reverse(d: dict) -> dict:
        d_rev = defaultdict(list)
        for k in d:
            for v in d[k]:
                d_rev[v].append(k)

        return dict(d_rev)

    def process(self,
                d_results: dict,
                reverse: bool) -> dict:
        d = {}

        for k in d_results:
            for ner in d_results[k]:
                d[k.upper().strip()] = [ner.upper().strip()]

        if reverse:
            return self._reverse(d)

        return d



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
""" Load the Ontology Model into Memory """


from baseblock import BaseObject
from rdflib import Graph

from owl_finder.singlequery.dmo import OwlGraphConnector


class LoadOntologyModel(BaseObject):
    """ Load the Ontology Model into Memory """

    def __init__(self,
                 ontology_name: str,
                 absolute_path: str,
                 prefix: str = None,
                 namespace: str = None):
        """ Change Log

        Created:
            25-May-2022
            craigtrim@gmail.com
            *   https://github.com/craigtrim/askowl/issues/1
        Updated:
            6-Jun-2022
            craigtrim@gmail.com
            *   refactored in pursuit of
                https://github.com/craigtrim/askowl/issues/2

        Args:
            ontology_name (str): the name of the Ontology
                used to infer the file path
            prefix (str, optional): the prefix. Defaults to None.
                assumed to the ontology_name unless specified otherwise
            namespace (str, optional): the model namespace. Defaults to None.
                assumed to be "http://graffl.ai/<ontology_name>" unless specified otherwise
            absolute_path (str): the absolute path to the OWL model
        """
        BaseObject.__init__(self, __name__)

        self._absolute_path = absolute_path
        self.ontology_name = self._get_ontology_name(ontology_name)
        self.prefix = self._get_prefix(prefix, ontology_name)
        self.namespace = self._get_namespace(namespace, ontology_name)

        if self.isEnabledForDebug:
            self.logger.debug('\n'.join([
                'Intialized Loader',
                f'\tPrefix: {self.prefix}',
                f'\tOntology Name: {self.ontology_name}',
                f'\tNamespace: {self.namespace}']))

    @staticmethod
    def _get_ontology_name(ontology_name: str) -> str:
        if not ontology_name.endswith('.owl'):
            return f'{ontology_name}.owl'
        return ontology_name

    @staticmethod
    def _get_prefix(prefix: str,
                    ontology_name: str) -> str:
        if not prefix:
            prefix = ontology_name
            if prefix.endswith('.owl'):
                prefix = prefix.split('.owl')[0].strip()
        return prefix

    @staticmethod
    def _get_namespace(namespace: str,
                       ontology_name: str) -> str:
        if not namespace:
            namespace = f'http://graffl.ai/{ontology_name}#'
        return namespace

    def process(self) -> Graph:

        dmo = OwlGraphConnector(
            prefix=self.prefix,
            namespace=self.namespace,
            ontology_name=self.ontology_name,
            absolute_path=self._absolute_path)

        return dmo.graph()



# !/usr/bin/env python
# -*- coding: UTF-8 -*-
""" Find spaCy NER (Named Entity Recognition) Types """


from baseblock import BaseObject


class FindNER(BaseObject):
    """ Find spaCy NER (Named Entity Recognition) Types """

    def __init__(self,
                 d_ner_depth: dict,
                 d_ner_taxo: dict,
                 d_spacy_ner: dict,
                 d_graffl_ner: dict):
        """ Change Log

        Created:
            13-Oct-2021
            craigtrim@gmail.com
            *   https://github.com/grafflr/graffl-core/issues/36
        Updated:
            14-Oct-2021
            craigtrim@gmail.com
            *   defect fixing updates
                https://github.com/grafflr/graffl-core/issues/42
        Updated:
            22-Oct-2021
            craigtrim@gmail.com
            *   replace manual depth dictionary with autogenerated values
                https://github.com/grafflr/graffl-core/issues/89#issuecomment-950061167
        Updated:
            27-Oct-2021
            craigtrim@gmail.com
            *   refactor functions into domain components in pursuit of
                https://github.com/grafflr/graffl-core/issues/94
        Updated:
            1-Feb-2022
            craigtrim@gmail.com
            *   make ontology param consistent; permit multiple values
                https://github.com/grafflr/graffl-core/issues/135#issuecomment-1027468040
        Updated:
            26-May-2022
            craigtrim@gmail.com
            *   treat 'ontologies' param as a list
                https://github.com/grafflr/deepnlu/issues/7
        Updated:
            27-May-2022
            craigtrim@gmail.com
            *   migrated to 'owlblock' in pursuit of
                https://github.com/grafflr/deepnlu/issues/13

        Args:
            ontologies (list): one-or-more Ontology models to use in processing
        """
        BaseObject.__init__(self, __name__)
        self._d_ner_depth = d_ner_depth
        self._d_ner_taxo = d_ner_taxo
        self._d_spacy_ner = d_spacy_ner
        self._d_graffl_ner = d_graffl_ner

    def _depth(self,
               ner: str) -> int or None:
        if ner in self._d_ner_depth:
            return self._d_ner_depth[ner][0]
        return None

    @staticmethod
    def _cleanse(input_text: str) -> str:
        if '_' in input_text:
            input_text = input_text.replace('_', ' ')
        return input_text

    def _find_spacy_ner(self,
                        input_text: str) -> list or None:

        # this is neither wrong nor unusual
        # it just means spaCy NERs were not defined in the OWL
        if not self._d_spacy_ner or not len(self._d_spacy_ner):
            return None

        input_text = self._cleanse(input_text)
        if input_text not in self._d_spacy_ner:
            return None

        return self._d_spacy_ner[input_text]

    def _find_graffl_ner(self,
                         input_text: str) -> list or None:

        # this is neither wrong nor unusual
        # it just means Graffl NERs were not defined in the OWL
        if not self._d_graffl_ner or not len(self._d_graffl_ner):
            return None

        input_text = self._cleanse(input_text)
        if input_text not in self._d_graffl_ner:
            return None

        return self._d_graffl_ner[input_text]

    def find_ner(self,
                 input_text: str) -> str or None:
        """Find NER from either the GRAFFL or spaCy taxonomy

        Reference:
            https://github.com/grafflr/graffl-core/issues/55#issuecomment-944572462

        Args:
            input_text (str): any input entity

        Returns:
            str or None: NER
        """
        input_text = self._cleanse(input_text)

        ners = []

        graffl_ners = self._find_graffl_ner(input_text)
        if graffl_ners:
            [ners.append(x) for x in graffl_ners]

        spacy_ners = self._find_spacy_ner(input_text)
        if spacy_ners:
            [ners.append(x) for x in spacy_ners]

        if not len(ners):
            ners.append('NER')

        if len(ners) == 1:
            return ners[0]

        d_depth = {self._depth(x): x for x in ners}
        return d_depth[max(d_depth)]



from .generate_view_spans import GenerateViewSpans
from .generate_view_synonyms import GenerateViewSynonyms
from .generate_view_trie import GenerateViewTrie
from .load_ontology_model import LoadOntologyModel
from .query_ontology_model import QueryOntologyModel
from .generate_plus_spans import GeneratePlusSpans



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
""" View Generator: Entity Spans for Long-Range Synonym Swapping """


from typing import List

from collections import defaultdict

from baseblock import BaseObject, EnvIO

# ----------------------------------------------------------
# Purpose:    List of Stopwords
# Reference:  https://github.com/craigtrim/askowl/issues/10
# Source:     https://gist.github.com/sebleier/554280#gistcomment-2596130
# Notes:      this list will likely be refined as needed over time
#             the 'stopwords' represented here are likely too coarse-grained
# ----------------------------------------------------------
stopwords = [
    'a', 'about', 'above', 'after', 'again', 'against', 'all', 'am', 'an', 'and', 'any', 'are', 'as', 'at', 'be', 'because', 'been', 'before',
    'between', 'both', 'but', 'by', 'can', 'did', 'do', 'does', 'doing', 'don', 'down', 'during', 'each', 'few', 'for', 'from', 'further',
    'had', 'has', 'have', 'having', 'he', 'her', 'here', 'hers', 'herself', 'him', 'himself', 'his', 'how', 'i', 'if', 'in', 'into', 'is', 'it', 'its', 'itself',
    'just', 'me', 'more', 'most', 'my', 'myself', 'no', 'nor', 'not', 'now', 'of', 'off', 'on', 'once', 'only', 'or', 'other', 'our', 'ours', 'ourselves', 'out',
    'over', 'own', 's', 'same', 'she', 'should', 'so', 'some', 'such', 't', 'than', 'that', 'the', 'their', 'theirs', 'them', 'themselves', 'then', 'there',
    'these', 'they', 'this', 'those', 'through', 'to', 'too', 'under', 'until', 'up', 'very', 'was', 'we', 'were', 'what', 'when', 'where', 'which', 'while',
    'who', 'whom', 'why', 'will', 'with', 'you', 'your', 'yours', 'yourself', 'yourselves'
]


class GenerateViewSpans(BaseObject):
    """ View Generator: Entity Spans for Long-Range Synonym Swapping """

    def __init__(self):
        """ Change Log

        Created:
            25-Oct-2021
            craigtrim@gmail.com
            *   Automate Long Distance Matching in Multi Gram Entities
        Updated:
            27-May-2022
            craigtrim@gmail.com
            *   ported to ask-owl
                https://github.com/craigtrim/askowl/issues/5
        Updated:
            24-Nov-2022
            craigtrim@gmail.com
            *   deal with CSV lists in rdfs:seeAlso relationship
                https://github.com/craigtrim/owl-finder/issues/2#issuecomment-1326808994
        Updated:
            28-Nov-2022
            craigtrim@gmail.com
            *   use hash to ensure uniqueness in spans
                https://github.com/craigtrim/owl-finder/issues/4#issuecomment-1329990189
        """
        BaseObject.__init__(self, __name__)

    @classmethod
    def _normalize(self,
                   d_results: dict) -> dict:
        """Normalize the Result of the OWL Query

        Sample Input
            'nurse_practitioner':   {   'Np':                       None,
                                        'Nps':                      None,
                                        'Nursing Practitioner':     None },
            'nursing':              {   'Nursing':                  None },
            'nursing_diploma':      {   'Nursing Diploma':          None },
            'nursing_education':    {   'Nursing Education':        None },

        Sample Output:
            {   'nurse_practitioner': [ 'Nursing Practitioner' ],
                'nursing_diploma    : []'
                'nursing_education' : []}

        Args:
            d_results (dict): [description]

        Returns:
            list: Terms to Span Over
        """
        d = {}

        def is_valid(term: str) -> bool:
            return ' ' in term or '_' in term

        keys = [k for k in d_results if is_valid(k)]

        for key in keys:

            def values() -> List[str]:
                """ Deal with CSV List

                Reference:
                    https://github.com/craigtrim/owl-finder/issues/2#issuecomment-1326808994

                Returns:
                    List[str]: a list of synonyms for this entity
                """
                s = set()
                for item in d_results[key]:
                    if ',' in item:
                        [s.add(y) for y in [
                            x.strip() for x in item.split(',')]
                         ]
                    else:
                        s.add(item)

                return sorted(s)

            values = [v for v in values() if is_valid(v)]
            values = [v.lower().replace(' ', '_') for v in values]
            values = [v for v in values if v != key]
            d[key] = values

        return d

    def _to_dict(self,
                 canon: str,
                 tokens: list) -> dict:

        def content() -> set:
            t = [x for x in tokens[1:] if x.lower() not in stopwords]
            return sorted(set(t), reverse=False)

        tokens2 = content()
        if not len(tokens2):
            return None

        # TODO: all these values have to be default only when not otherwise specified
        distance = EnvIO.int_or_default('SPAN_DISTANCE', 3)
        return {
            'content': content(),
            'distance': distance,
            'forward': True,
            'reverse': True,
            'canon': canon
        }

    def process(self,
                d_results: dict) -> dict:

        d = defaultdict(list)

        hashes = set()
        d_results = self._normalize(d_results)

        def update(canon: str,
                   term: str) -> None:
            tokens = term.split('_')
            trigger = tokens[0]

            d_result = self._to_dict(canon, tokens)

            if d_result:

                # 20221128; Ensure Uniqueness
                # https://github.com/craigtrim/owl-finder/issues/4#issuecomment-1329990189
                hash_d_result = hash(''.join([
                    str(x) for x in d_result.values()
                ]))

                if hash_d_result not in hashes:
                    d[trigger].append(d_result)
                    hashes.add(hash_d_result)

        for k in d_results:
            update(canon=k, term=k)
            for v in d_results[k]:
                update(canon=k, term=v)

        return dict(d)



# !/usr/bin/env python
# -*- coding: UTF-8 -*-
""" Find a Pallete to associate to NER types """


import os
from random import randint

from baseblock import BaseObject, FileIO


class NerPalleteLookup(BaseObject):
    """ Find a Pallete to associate to NER types """

    _d_merge_taxorev = None

    def __init__(self,
                 d_ner_taxonomy_rev: dict):
        """ Change Log

        Created:
            27-Oct-2021
            craigtrim@gmail.com
            *   https://github.com/grafflr/graffl-core/issues/94
        Updated:
            1-Feb-2022
            craigtrim@gmail.com
            *   enforce ontologies as a list param in domain components
                https://github.com/grafflr/graffl-core/issues/135#issuecomment-1027464370
        Updated:
            27-May-2022
            craigtrim@gmail.com
            *   migrated into 'owlblock' in pursuit of
                https://github.com/grafflr/deepnlu/issues/13

        Args:
            d_ner_taxonomy_rev (dict): reversed NER Taxonomy dictionary
        """
        BaseObject.__init__(self, __name__)
        self._d_palletes = self._load_config()

        self._d_ner_taxonomy_rev = d_ner_taxonomy_rev
        self._ner_taxonomy_keys = sorted(self._d_ner_taxonomy_rev.keys())

        self._d_assoc = self._configure()

    @staticmethod
    def _load_config() -> dict:
        path = os.path.normpath(os.path.join(
            os.getcwd(),
            'resources/palletes.yaml'))

        FileIO.exists_or_error(path)

        return FileIO.read_yaml(path)['palletes']

    def _configure(self) -> int or None:

        d_assoc = {}
        used_keys = []

        pallete_keys = sorted(self._d_palletes.keys())

        for i in range(len(self._ner_taxonomy_keys)):

            def pallete_key(n: int) -> int:
                x = randint(0, len(pallete_keys) - 1)
                if x not in used_keys:
                    used_keys.append(x)
                    return x
                if n > len(pallete_keys):
                    return 0
                return pallete_key(n + 1)

            ners = self._d_ner_taxonomy_rev[self._ner_taxonomy_keys[i]]
            pallete = pallete_keys[pallete_key(0)]
            colors = self._d_palletes[pallete]['colors']

            for j in range(len(ners)):
                if ners[j] not in d_assoc:

                    def color_key() -> int:
                        if j >= len(colors):
                            return 0
                        return j

                    d_assoc[ners[j]] = f'#{colors[color_key()]}'

        return d_assoc

    def lookup(self,
               input_text: str) -> str:
        return self._d_assoc[input_text]

    def colors(self) -> dict:
        return self._d_assoc



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
""" Normalize the Extracted OWL Query Data """


from baseblock import BaseObject
from rdflib import Graph


class OwlQueryNormalize(BaseObject):
    """ Normalize the Extracted OWL Query Data """

    def __init__(self,
                 graph: Graph):
        """ Change Log

        Created:
            13-Oct-2021
            craigtrim@gmail.com
            *   refactored out of 'owl-data-extract'
                Build Owl2PY dictionary for backwardCompatibilityTypes
        Updated:
            25-May-2022
            craigtrim@gmail.com
            *   refactor into 'askowl' repo
                https://github.com/craigtrim/askowl/issues/1

        Args:
            graph (Graph): an instantiated RDF graph
        """
        BaseObject.__init__(self, __name__)
        self._graph = graph

    def process(self,
                d_results: dict,
                all_subjects: bool = True) -> dict:

        d_normalized = {}

        # Represent all Subjects
        if all_subjects:
            for subject, _, _ in self._graph:
                subject = subject.title().split('#')[-1].strip().lower()
                if subject.startswith('http'):
                    continue  # this is the root entity
                d_normalized[subject] = set()

        # Add Query Results
        for k in d_results:
            d_normalized[k] = d_results[k]

        # # Sort Values
        d_normalized = {k: d_normalized[k] for k in d_normalized
                        if len(d_normalized[k])}

        return d_normalized



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
""" View Generator: Trie Grams """


from collections import defaultdict

from baseblock import BaseObject


class GenerateViewTrie(BaseObject):
    """ View Generator: Trie Grams """

    def __init__(self):
        """ Change Log

        Created:
            27-May-2022
            craigtrim@gmail.com
            *   https://github.com/craigtrim/askowl/issues/8
        """
        BaseObject.__init__(self, __name__)

    @staticmethod
    def _to_gramsize_dict(d_results: dict) -> dict:
        d = defaultdict(list)
        for k in d_results:

            name = k.replace('_', ' ')
            gram_size = len(name.split(' '))
            d[gram_size].append(name)

        return d

    @staticmethod
    def _sort(tokens: list) -> list:
        values = sorted(set(tokens), key=len)
        values.reverse()
        return values

    def _trie(self,
              tokens: list) -> dict:
        d = defaultdict(list)

        for k in tokens:
            values = k.split(' ')
            d[values[0]].append(' '.join(values[1:]))

        def decide(obj):

            if type(obj) == list:
                if ' ' in obj[0]:
                    return self._trie(obj)
                return self._sort(obj)
            raise NotImplementedError(type(obj))

        return {k: decide(d[k]) for k in d}

    def process(self,
                d_results: dict) -> dict:

        d_gramsize = self._to_gramsize_dict(d_results)

        d_gramsize[1] = self._sort(d_gramsize[1])
        d_gramsize[2] = self._trie(d_gramsize[2])
        d_gramsize[3] = self._trie(d_gramsize[3])
        d_gramsize[4] = self._trie(d_gramsize[4])
        d_gramsize[5] = self._trie(d_gramsize[5])
        d_gramsize[6] = self._trie(d_gramsize[6])

        return dict(d_gramsize)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
""" Perform the RDF Query """


from collections import defaultdict

from baseblock import BaseObject
from rdflib import Graph, Literal, URIRef
from rdflib.plugins.sparql.processor import SPARQLResult

from owl_finder.singlequery.dto import QueryResultType


class OwlQueryExtract(BaseObject):
    """ Perform the RDF Query """

    def __init__(self,
                 graph: Graph):
        """ Change Log

        Created:
            13-Oct-2021
            craigtrim@gmail.com
            *   refactored out of 'owl-data-extract'
                Build Owl2PY dictionary for backwardCompatibilityTypes
        Updated:
            2-Feb-2022
            craigtrim@gmail.com
            *   do not split designated string datatypes
                https://github.com/craigtrim/askowl/issues/3
        Updated:
            25-May-2022
            craigtrim@gmail.com
            *   refactor into 'ask-owl' repo
                https://github.com/craigtrim/askowl/issues/1
        Updated:
            27-May-2022
            craigtrim@gmail.com
            *   expose ability to lowercase output
                https://github.com/craigtrim/askowl/issues/4

        Args:
            graph (Graph): an instantiated RDF graph
        """
        BaseObject.__init__(self, __name__)
        self._graph = graph

    def _log_no_rows_found(self,
                           query_type: QueryResultType) -> None:
        self.logger.debug('\n'.join([
            'Result Set Transformation Failure',
            '\tReason: No Rows Found',
            f'\tTransform: {query_type.name}']))

    def _transform(self,
                   value: object,
                   to_lowercase: bool) -> str:
        _type = type(value)

        if _type == URIRef:
            value = value.title().split('#')[-1].strip()
            if to_lowercase:
                value = value.lower()
            return value

        elif _type == Literal:
            value = str(value).strip()
            if to_lowercase:
                value = value.lower()
            return value

        else:
            self.logger.error('\n'.join([
                'DataType Not Recognized',
                f'\tActual Type: {_type}']))
            raise NotImplementedError

    def _list_of_strings(self,
                         query_results: SPARQLResult,
                         to_lowercase: bool) -> list:
        results = []

        rows = [x for x in query_results if len(x) >= 1]

        if not len(rows):
            self._log_no_rows_found(QueryResultType.LIST_OF_STRINGS)
            return None

        for row in rows:
            results.append(self._transform(row[0], to_lowercase))

        return results

    def _dict_of_str2str(self,
                         query_results: SPARQLResult,
                         to_lowercase: bool) -> dict:
        d = {}

        rows = [x for x in query_results if len(x) >= 2]
        if not len(rows):
            self._log_no_rows_found(QueryResultType.DICT_OF_STR2STR)
            return None

        for row in rows:
            key = self._transform(row[0], to_lowercase)
            value = self._transform(row[1], to_lowercase)
            d[key] = value

        return d

    def _dict_of_str2list(self,
                          query_results: SPARQLResult,
                          to_lowercase: bool) -> dict:
        d = defaultdict(list)

        rows = [x for x in query_results if len(x) >= 2]
        if not len(rows):
            self._log_no_rows_found(QueryResultType.DICT_OF_STR2LIST)
            return None

        for row in rows:
            key = self._transform(row[0], to_lowercase)
            value = self._transform(row[1], to_lowercase)
            d[key].append(value)

        return dict(d)

    def _update(self,
                query_results: SPARQLResult,
                to_lowercase: bool,
                result_type: QueryResultType) -> object:

        if result_type == QueryResultType.LIST_OF_STRINGS:
            return self._list_of_strings(query_results, to_lowercase)

        elif result_type == QueryResultType.DICT_OF_STR2STR:
            return self._dict_of_str2str(query_results, to_lowercase)

        elif result_type == QueryResultType.DICT_OF_STR2LIST:
            return self._dict_of_str2list(query_results, to_lowercase)

        else:
            raise NotImplementedError

    def process(self,
                query: str,
                to_lowercase: bool,
                result_type: QueryResultType) -> dict:
        try:

            result = self._graph.query(query)
            if result_type == QueryResultType.DO_NOT_TRANSFORM:
                return result

            svcresult = self._update(
                query_results=result,
                to_lowercase=to_lowercase,
                result_type=result_type)

            return svcresult
        except Exception as e:
            print(e)
            self.logger.error('\n'.join([
                'Parsing Exception',
                f'\tQuery:\n{query}']))
            raise ValueError('OWL Query Failed')



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
""" View Generator: Lookup Spans using N-Gram Input """


from baseblock import BaseObject, Stopwatch

from owl_finder.multiquery.dmo.span import OwlSpanAugment, OwlSpanGenerate


class ViewGeneratorLookup(BaseObject):
    """ View Generator: Lookup Spans using N-Gram Input """

    def __init__(self):
        """ Change Log

        Created:
            26-May-2022
            craigtrim@gmail.com
            *   https://github.com/grafflr/ask-owl/issues/4
        Created:
            30-May-2022
            craigtrim@gmail.com
            *   migrated from 'askowl' into 'deepnlu'
                https://github.com/grafflr/deepnlu/issues/21#issuecomment-1141518333
        """
        BaseObject.__init__(self, __name__)

    @staticmethod
    def _sort(tokens: list) -> list:
        values = sorted(set(tokens), key=len)
        values.reverse()
        return values

    def process(self,
                d_results: dict) -> dict:

        sw = Stopwatch()

        d_results = OwlSpanGenerate().process(d_results)
        d_results = OwlSpanAugment().process(d_results)

        d_results[1] = self._sort(d_results[1])
        d_results[2] = self._sort(d_results[2])
        d_results[3] = self._sort(d_results[3])
        d_results[4] = self._sort(d_results[4])
        d_results[5] = self._sort(d_results[5])
        d_results[6] = self._sort(d_results[6])

        if self.isEnabledForInfo:
            self.logger.info('\n'.join([
                'Generated OWL Spans',
                f'\tTotal Time: {str(sw)}',
                f'\tTotal Size: {len(d_results)}']))

        return d_results



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
""" View Generator: Graffl NER Types """


from collections import defaultdict

from baseblock import BaseObject


class ViewGeneratorNerDepth(BaseObject):
    """ View Generator: Graffl NER Types """

    def __init__(self):
        """ Change Log

        Created:
            12-Oct-2021
            craig.@graffl.ai
            *   https://github.com/grafflr/graffl-core/issues/38
        Updated:
            15-Oct-2021
            craig.@graffl.ai
            *   renamed from 'spacy-ner-transform'
                https://github.com/grafflr/graffl-core/issues/55
        Updated:
            27-May-2022
            craigtrim@gmail.com
            *   ported to 'deepnlu'
                https://github.com/grafflr/deepnlu/issues/10
        """
        BaseObject.__init__(self, __name__)

    @staticmethod
    def _reverse(d: dict) -> dict:
        d_rev = defaultdict(list)
        for k in d:
            for v in d[k]:
                d_rev[v].append(k)

        return dict(d_rev)

    def process(self,
                d_results: dict,
                reverse: bool) -> dict:
        """Find the depth of each NER entity to root

        The higher the associated number, the more specific the NER is

        Args:
            d_results (dict): [description]
            Sample Format:
                [
                    'competency':   {'2': 'integer'},
                    'condition':    {'0': 'integer'},
                    'continent':    {'1': 'integer'},
                ]

        Returns:
            dict: owl2py dictionary
        """
        d = {
            'NER': '0'  # default
        }

        for k in d_results:
            depth = int(d_results[k][0])
            d[k.upper()] = str(depth + 1)

        if reverse:
            return self._reverse(d)

        return dict(d)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
""" Load a Supplementary External File of Synonyms """


import logging

from baseblock import BaseObject, FileIO, Stopwatch


class ExternalSynonymLoader(BaseObject):
    """ Load a Supplementary External File of Synonyms """

    def __init__(self,
                 synonym_file_path: str):
        """ Change Log

        Created:
            8-Oct-2021
            craigtrim@gmail.com
            *   https://github.com/grafflr/graffl-core/issues/13#issuecomment-939103068
        Updated:
            29-May-2022
            craigtrim@gmail.com
            *   migrated out of graffl-core in pursuit of
                https://github.com/grafflr/deepnlu/issues/18

        Args:
            synonym_file_path (str): absolute (and verified) path to the external synonyms file
        """
        BaseObject.__init__(self, __name__)
        self._synonym_file_path = synonym_file_path

    def _process(self) -> dict:
        d_file = {}

        def is_valid(line: str) -> bool:
            if not len(line):
                return False
            if line.startswith('#'):
                return False
            if '~' not in line:
                return False
            return True

        lines = FileIO.read_lines(self._synonym_file_path)
        lines = [x.strip() for x in lines if x]
        lines = [x for x in lines if is_valid(x)]

        for line in lines:
            tokens = line.lower().split('~')
            canon = tokens[0]

            variants = tokens[-1].split(',')
            variants = [x.strip() for x in variants if x]
            variants = sorted(set([x for x in variants if len(x)]), key=len)
            variants.reverse()

            d_file[canon] = variants

        return d_file

    def process(self) -> dict:
        sw = Stopwatch()

        d_file = self._process()

        if self.logger.isEnabledFor(logging.DEBUG):
            self.logger.debug('\n'.join([
                'Loaded External File',
                f'\tFile Path: {self._synonym_file_path}',
                f'\tTotal Size: {len(d_file)}',
                f'\tTotal Time: {str(sw)}']))

        return d_file



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
""" API for the ask-owl Microservice """


from pprint import pprint
from functools import lru_cache
from collections import defaultdict

from baseblock import Enforcer
from baseblock import BaseObject

from rdflib.plugins.sparql.processor import SPARQLResult

from owl_finder.singlequery.dto import QueryResultType
from owl_finder.singlequery.svc import GeneratePlusSpans
from owl_finder.singlequery.svc import GenerateViewSpans
from owl_finder.singlequery.svc import GenerateViewSynonyms
from owl_finder.singlequery.svc import GenerateViewTrie
from owl_finder.singlequery.svc import LoadOntologyModel
from owl_finder.singlequery.svc import QueryOntologyModel


class AskOwlAPI(BaseObject):
    """ API for the ask-owl Microservice """

    def __init__(self,
                 ontology_name: str,
                 absolute_path: str):
        """ Change Log

        Created:
            25-May-2022
            craigtrim@gmail.com
            *   https://github.com/craigtrim/askowl/issues/1
        Created:
            6-Jun-2022
            craigtrim@gmail.com
            *   simplify __init__ method in pursuit of
                https://github.com/craigtrim/askowl/issues/2
        Updated:
            25-Nov-2022
            craigtrim@gmail.com
            *   add generate-plus-spans
                https://github.com/craigtrim/owl-finder/issues/4

        Args:
            ontology_name (str): the name of the Ontology (OWL) model
            absolute_path (str): the absolute path to the OWL model

        Raises:
            ValueError: OWL Model Not Found
        """
        BaseObject.__init__(self, __name__)
        self.ontology_name = ontology_name
        self.absolute_path = absolute_path

        self.graph, self.prefix, self.namespace = self._load_models(
            ontology_name=ontology_name,
            absolute_path=absolute_path)

        self._execute_query = QueryOntologyModel(self.graph).process

    def _load_models(self,
                     ontology_name: str,
                     absolute_path: str):
        try:

            dmo = LoadOntologyModel(
                ontology_name=ontology_name,
                absolute_path=absolute_path)

            return dmo.process(), dmo.prefix, dmo.namespace

        except FileNotFoundError as e:
            self.logger.error('\n'.join([
                'OWL Model Not Found',
                f'\tName: {ontology_name}',
                f'\tAbsolute Path: {absolute_path}', e]))
            raise FileNotFoundError(absolute_path)

        # I don't know why the IDE considers this code unreachable ...
        raise ValueError

    def adhoc(self,
              sparql_query: str,
              to_lowercase: bool,
              result_type: QueryResultType = QueryResultType.DO_NOT_TRANSFORM) -> SPARQLResult:
        """ Execute an ad-hoc SPARQL query on the OWL model

        Args:
            sparql_query (str): the SPARQL query
                no validation will occur on this query

        Returns:
            SPARQLResult: the SPARQL result set
        """
        return self._execute_query(
            reverse=False,
            sparql=sparql_query,
            result_type=result_type,
            to_lowercase=to_lowercase)

    @lru_cache(maxsize=6, typed=False)
    def ngrams(self,
               gram_level: 1) -> list or None:
        """ Generate n-Grams by Size

        Args:
            gram_level (1): the gram level to query for

        Returns:
            list or None: the n-Gram results (if any)
        """
        sparql = 'SELECT ?a WHERE { ?a rdfs:subClassOf ?b }'

        results = self._execute_query(
            sparql=sparql,
            to_lowercase=True,
            result_type=QueryResultType.LIST_OF_STRINGS,
        )

        if not results or not len(results):
            return None

        return [x for x in results if x.count('_') == gram_level - 1]

    @lru_cache
    def trie(self) -> dict or None:
        """ Generate Entities in a Trie View

        Sample Input:
            ['First Quarter Results', 'First Quarter GDP', 'First Time']

        Sample Output:
            {
                'First': {
                    'Quarter': ['Results', 'GDP'],
                    'Time': [],
                }
            }

        Reference:
            https://github.com/craigtrim/askowl/issues/8

        Args:
            to_lowercase (bool, optional): lowercase all data. Defaults to True.

        Returns:
            dict: dictionary of values keyed by n-gram size
        """
        sparql = 'SELECT ?a ?b WHERE  {  ?a rdfs:subClassOf ?b }'
        d_results = self._execute_query(
            sparql=sparql,
            to_lowercase=True,
            result_type=QueryResultType.DICT_OF_STR2LIST,
        )

        if not d_results or not len(d_results):
            return None

        return GenerateViewTrie().process(d_results)

    @lru_cache(typed=False)
    def by_subject_predicate(self,
                             predicate: str,
                             subject: str,
                             to_lowercase: bool = True) -> list:
        """ Retrieve a list of values by custom subject and custom predicate

        Args:
            predicate (str): the predicate name
            subject (str): the subject name
            to_lowercase (str, optional): optionally lowercase all results.  Defaults to True.

        Returns:
            list: a list of values for this predicate
        """
        sparql = 'SELECT ?b WHERE { #PREFIX:#SUBJECT rdf:type owl:Class ; #PREFIX:#PREDICATE ?b }'
        sparql = sparql.replace('#PREFIX', self.prefix)
        sparql = sparql.replace('#PREDICATE', predicate)
        sparql = sparql.replace('#SUBJECT', subject)

        d_results = self._execute_query(
            sparql=sparql,
            to_lowercase=to_lowercase,
            result_type=QueryResultType.LIST_OF_STRINGS,
        )

        return d_results

    @lru_cache(typed=False)
    def by_predicate(self,
                     predicate: str,
                     to_lowercase: bool = True,
                     reverse: bool = False) -> list:
        """ Retrieve a list of values by custom predicate

        Args:
            predicate (str): the predicate name
            to_lowercase (str, optional): optionally lowercase all results.  Defaults to True.
            reverse (bool, optional): reverses the subject/object order. Defaults to False.
                if "?x implies ?y" and reverse=False
                    the results will be { ?x: [?y-1, ?y-2, ..., ?y-N]}
                if "?x implies ?y" and reverse=True
                    the results will be { ?y-1: [x], ?y-2: [x], ?y-N: [x]}

        Returns:
            dict: a dict of values for this predicate
        """
        sparql = 'SELECT ?a ?b WHERE { ?a #PREFIX:#PREDICATE ?b }'

        def get_prefix() -> str:
            if ':' in predicate:
                return predicate.split(':')[0].strip()
            return self.prefix

        def get_predicate() -> str:
            if ':' in predicate:
                return predicate.split(':')[-1].strip()
            return predicate

        _prefix = get_prefix()
        _predicate = get_predicate()

        sparql = sparql.replace('#PREFIX', _prefix)
        sparql = sparql.replace('#PREDICATE', _predicate)

        d_results = self._execute_query(
            sparql=sparql,
            reverse=reverse,
            to_lowercase=to_lowercase,
            result_type=QueryResultType.DICT_OF_STR2LIST,
        )

        return d_results

    @lru_cache
    def labels(self) -> list:
        """ Retrieve rdfs:label values from the Graph

        Returns:
            list: a list of labels
        """
        sparql = 'SELECT ?a WHERE { ?x rdfs:label ?a }'
        return self._execute_query(
            sparql=sparql,
            to_lowercase=False,
            result_type=QueryResultType.LIST_OF_STRINGS)

    @lru_cache
    def comments(self) -> list:
        """ Retrieve rdfs:comment values from the Graph

        Returns:
            list: a list of labels
        """
        sparql = 'SELECT ?a ?b WHERE { ?a rdfs:comment ?b }'

        d_results = self._execute_query(
            sparql=sparql,
            to_lowercase=False,
            result_type=QueryResultType.DICT_OF_STR2LIST,
        )

        return d_results

    @lru_cache
    def see_also(self,
                 to_lowercase: bool = True) -> list:
        """ Retrieve rdfs:seeAlso values from the Graph

        Args:
            to_lowercase (str, optional): optionally lowercase all results.  Defaults to True.

        Returns:
            list: a list of see-also values
        """
        sparql = 'SELECT ?a WHERE { ?x rdfs:seeAlso ?a }'
        return self._execute_query(
            sparql=sparql,
            to_lowercase=to_lowercase,
            result_type=QueryResultType.LIST_OF_STRINGS)

    @lru_cache
    def backward_compatible_with(self,
                                 to_lowercase: bool = True) -> list:
        """ Retrieve owl:backwardCompatibleWith values from the Graph

        Args:
            to_lowercase (str, optional): optionally lowercase all results.  Defaults to True.

        Returns:
            list: a list of see-also values
        """
        sparql = 'SELECT ?a WHERE { ?x rdfs:seeAlso ?a }'
        return self._execute_query(
            sparql=sparql,
            to_lowercase=to_lowercase,
            result_type=QueryResultType.LIST_OF_STRINGS)

    @lru_cache
    def types(self,
              to_lowercase: bool = True) -> list:
        """ Retrieve rdf:type values from the Graph

        Args:
            to_lowercase (str, optional): optionally lowercase all results.  Defaults to True.

        Returns:
            list: a list of labels
        """
        sparql = 'SELECT ?a WHERE { ?a rdf:type ?x }'
        return self._execute_query(
            sparql=sparql,
            to_lowercase=to_lowercase,
            result_type=QueryResultType.LIST_OF_STRINGS)

    @lru_cache
    def _synonym_query(self) -> dict or None:
        """ Generate n-Gram Spans suitable for Synonym Matching

        Reference:
            https://github.com/craigtrim/askowl/issues/8

        Returns:
            dict: dictionary of values keyed by n-gram size
        """
        sparql = """
        SELECT
            ?a ?b
        WHERE
        {
            {
                { ?a rdfs:label ?b }
                OPTIONAL {?a rdfs:seeAlso ?b}
            } UNION
            {
                { ?a rdfs:seeAlso ?b }
                OPTIONAL {?a rdfs:label ?b}

            }
        }
        """
        d_results = self._execute_query(
            sparql=sparql,
            to_lowercase=True,
            result_type=QueryResultType.DICT_OF_STR2LIST,
        )

        if not d_results or not len(d_results):
            return None

        d_normalized = defaultdict(list)
        for k in d_results:
            for synonym in d_results[k]:
                synonyms = [x.strip() for x in synonym.split(',')]
                synonyms = [x for x in synonyms if x and len(x)]
                [d_normalized[k].append(x) for x in synonyms]

        return d_normalized

    @lru_cache
    def synonyms(self) -> dict or None:
        """ Generate a Dictionary of Entities keyed to Synonym Lists

        Reference:
            https://github.com/craigtrim/askowl/issues/8

        Returns:
            dict: dictionary of entities keyed by synonym
        """
        d_results = self._synonym_query()

        if not d_results or not len(d_results):
            return None

        return GenerateViewSynonyms().process(d_results)

    @lru_cache
    def synonyms_rev(self) -> dict:
        """ Reverse Synonym Dictionary:
        Synonyms are keyed to one-or-more Entities

        Reference:
            https://github.com/craigtrim/askowl/issues/8

        Returns:
            dict: dictionary of synonyms keyed by entity
        """
        d_results = self._synonym_query()

        if not d_results or not len(d_results):
            return None

        return GenerateViewSynonyms().process(d_results, reverse=True)

    @lru_cache
    def spans(self) -> dict or None:
        """ Entity Spans for Long-Range Matching

        Reference:
            https://github.com/craigtrim/askowl/issues/5

        Returns:
            dict: dictionary of spans keyed by entity name
        """
        sparql = 'SELECT ?a ?b WHERE { { ?a rdfs:label ?b } UNION { ?a rdfs:seeAlso ?b } }'

        d_results = self._execute_query(
            sparql=sparql,
            to_lowercase=True,
            result_type=QueryResultType.DICT_OF_STR2LIST,
        )

        if not d_results or not len(d_results):
            return None

        d_merged = defaultdict(list)

        def merge(d: dict) -> None:
            if not d:
                return None

            # Change Log:
            # https://github.com/craigtrim/owl-parser/issues/7
            # Ensuring output dict is str:list[str]
            for k in d:
                [d_merged[k].append(x) for x in d[k]]

        merge(GeneratePlusSpans().process(d_results))
        merge(GenerateViewSpans().process(d_results))

        d_merged = dict(d_merged)

        if self.isEnabledForDebug:
            Enforcer.is_dict_of_list_of_dicts(d_merged)

        return d_merged



# !/usr/bin/env python
# -*- coding: UTF-8 -*-
""" Generic Facade to Find Data in 1..* Ontology Models """


from baseblock import BaseObject

from owl_finder.multiquery.dmo import ModelResultMerge, ViewGeneratorNerLabel
from owl_finder.singlequery.dto import QueryResultType


class QueryNerLabel(BaseObject):
    """ Generic Facade to Find Data in 1..* Ontology Models """

    __SPARQL_QUERY = """
            SELECT
                ?label ?NER
            WHERE
            {
                ?entity owl:backwardCompatibleWith ?NER
                {
                    ?child rdfs:subClassOf* ?entity
                    { ?child rdfs:label ?label }
                    UNION
                    { ?child rdfs:seeAlso ?label }
                }
                OPTIONAL
                {
                    { ?entity rdfs:label ?label }
                    UNION
                    { ?entity rdfs:seeAlso ?label }
                }

                FILTER
                (
                    datatype(?NER) = $PREFIX:$NER
                )
            }
        """

    def __init__(self,
                 d_ontologies: list):
        """ Change Log

        Created:
            27-May-2022
            craigtrim@gmail.com
            *   https://github.com/grafflr/deepnlu/issues/13

        Args:
            d_ontologies (list): in-memory graph model keyed by name
        """
        BaseObject.__init__(self, __name__)
        self._merge = ModelResultMerge().process
        self._d_ontologies = d_ontologies
        self._generate_view = ViewGeneratorNerLabel().process

    def _get_results(self,
                     ner_type: str) -> list:
        results = []
        for ontology_name in self._d_ontologies:

            ask_owl_api = self._d_ontologies[ontology_name]

            sparql_query = self.__SPARQL_QUERY

            sparql_query = sparql_query.replace(
                '$PREFIX', ontology_name)

            sparql_query = sparql_query.replace(
                '$NER', ner_type)

            results.append(ask_owl_api.adhoc(
                sparql_query=sparql_query,
                to_lowercase=True,
                result_type=QueryResultType.DICT_OF_STR2LIST))

        if not results:
            return None

        elif len(results) == 1:
            return results[0]

        return self._merge(results, QueryResultType.DICT_OF_STR2LIST)

    def process(self,
                ner_type: str,
                reverse: bool = False) -> dict or None:

        d_results = self._get_results(ner_type)

        if not d_results or not len(d_results):
            return None

        return self._generate_view(
            reverse=reverse,
            d_results=d_results)



# !/usr/bin/env python
# -*- coding: UTF-8 -*-
""" Generic Facade to find Synonym Data on Disk """

import os
from collections import defaultdict
from functools import lru_cache
from pprint import pprint
from typing import Callable

from baseblock import BaseObject, Enforcer, FileIO

from owl_finder.multiquery.dmo import ExternalSynonymLoader
from owl_finder.singlequery import AskOwlAPI
from owl_finder.singlequery.dto import QueryResultType


class LoadSynonyms(BaseObject):
    """ Generic Facade to find Synonym Data on Disk """

    def __init__(self,
                 d_ontologies: dict,
                 model_result_merge: Callable):
        """ Change Log

        Created:
            29-May-2022
            craigtrim@gmail.com
            *   in pursuit of
                https://github.com/grafflr/deepnlu/issues/18

        Args:
            d_ontologies (dict): a key:value dictionary of 'ontology_name<str> : AskOwlAPI()'
            model_result_merge (Callable): a callback to the model-result-merge component
        """
        BaseObject.__init__(self, __name__)
        self._merge = model_result_merge
        self._d_ontologies = d_ontologies

        self._d_external_fwd, self._d_external_rev = self._external_synonyms()

    def _external_synonyms(self) -> tuple:
        """ Load External Synonyms from File

        It's common for an ontology to have a complementary text file
            with entity names keyed by a list of synonyms

        The file name and location are identical to the OWL model
            with the exception of the file extension (.txt instead of .owl)

        Returns:
            _type_: a dictionary of external synonyms
        """

        d_external_fwd = {}
        d_external_rev = {}

        for ontology_name in self._d_ontologies:

            def api_instance() -> AskOwlAPI:
                return self._d_ontologies[ontology_name]

            ask_owl_api = api_instance()

            synonym_file_path = os.path.normpath(os.path.join(
                ask_owl_api.absolute_path,
                f'{ask_owl_api.ontology_name}.txt'))

            def get_synonyms() -> dict:
                if FileIO.exists(synonym_file_path):
                    return ExternalSynonymLoader(synonym_file_path).process()
                return {}  # the file is optional

            def get_reverse(d_fwd: dict) -> dict:
                d_rev = defaultdict(list)
                for k in d_fwd:
                    for v in d_fwd[k]:
                        d_rev[v].append(k)
                return dict(d_rev)

            d_fwd = get_synonyms()
            d_external_fwd[ontology_name] = d_fwd
            d_external_rev[ontology_name] = get_reverse(d_fwd)

        return d_external_fwd, d_external_rev

    @lru_cache
    def synonyms(self) -> dict:
        results = []
        for ontology_name in self._d_ontologies:

            d_owl_synonyms = self._d_ontologies[ontology_name].synonyms()
            if d_owl_synonyms and len(d_owl_synonyms):
                results.append(d_owl_synonyms)

            d_txt_synonyms = self._d_external_fwd[ontology_name]
            if d_txt_synonyms and len(d_txt_synonyms):
                results.append(d_txt_synonyms)

        if not results or not len(results):
            return None

        return self._merge(results, QueryResultType.DICT_OF_STR2LIST)

    @lru_cache
    def synonyms_rev(self) -> dict:
        results = []
        for ontology_name in self._d_ontologies:

            d_owl_synonyms = self._d_ontologies[ontology_name].synonyms_rev()
            if d_owl_synonyms and len(d_owl_synonyms):
                results.append(d_owl_synonyms)

            d_txt_synonyms = self._d_external_rev[ontology_name]
            if d_txt_synonyms and len(d_txt_synonyms):
                results.append(d_txt_synonyms)

        if not results:
            return None
        # elif len(results) == 1:
        #     return results[0]

        return self._merge(results, QueryResultType.DICT_OF_STR2LIST)



# !/usr/bin/env python
# -*- coding: UTF-8 -*-
""" Generic Facade to find Synonym Data on Disk """

from pprint import pprint

from baseblock import BaseObject, Enforcer


class FindSynonyms(BaseObject):

    """ Generic Facade to find Synonym Data on Disk """

    def __init__(self,
                 d_synonyms_fwd: dict,
                 d_synonyms_rev: dict):
        """ Change Log

        Created:
            7-Oct-2021
            craigtrim@gmail.com
            *   https://github.com/grafflr/graffl-core/issues/8
        Updated:
            11-Oct-2021
            craigtrim@gmail.com
            *   support args for 1..* names in param
                https://github.com/grafflr/graffl-core/issues/30#issuecomment-940252993
        Updated:
            25-Jan-2022
            craigtrim@gmail.com
            *   pass ontology-name as optional param
                https://github.com/grafflr/graffl-core/issues/135
        Updated:
            1-Feb-2022
            craigtrim@gmail.com
            *   use baseblock ontology name loader
                https://github.com/grafflr/graffl-core/issues/135
            *   a finder initialization is a contract
                https://github.com/grafflr/graffl-core/issues/135#issuecomment-1027474785
        Updated:
            26-May-2022
            craigtrim@gmail.com
            *   treat 'ontologies' param as a list
                https://github.com/grafflr/deepnlu/issues/7
        Updated:
            27-May-2022
            craigtrim@gmail.com
            *   migrated to 'owlblock' in pursuit of
                https://github.com/grafflr/deepnlu/issues/13
        Updated:
            30-May-2022
            craigtrim@gmail.com
            *   resilience testing
                https://github.com/grafflr/deepnlu/issues/20

        Args:
            ontologies (list): one-or-more Ontology models to use in processing
        """
        BaseObject.__init__(self, __name__)
        self._d_synonyms_fwd = d_synonyms_fwd
        self._d_synonyms_rev = d_synonyms_rev

    def d_synonyms_fwd(self):
        return self._d_synonyms_fwd

    def d_synonyms_rev(self):
        return self._d_synonyms_rev

    def _cleanse_canon(self,
                       input_text: str) -> str:
        if ' ' in input_text:
            input_text = input_text.replace(' ', '_')
        return input_text.lower().strip()

    def _cleanse_variant(self,
                         input_text: str) -> str:
        if '_' in input_text:
            input_text = input_text.replace('_', ' ')
        return input_text.lower().strip()

    def _has_synonyms_fwd(self) -> bool:
        """ No Synonyms Exist
        This result is unusual outside of unit-testing; but could happen

        Reference:
            https://github.com/grafflr/deepnlu/issues/20

        Returns:
            bool: True if synonyms exist
        """
        return self._d_synonyms_fwd and len(self._d_synonyms_fwd)

    def _has_synonyms_rev(self) -> bool:
        """ No Synonyms Exist
        This result is unusual outside of unit-testing; but could happen

        Reference:
            https://github.com/grafflr/deepnlu/issues/20

        Returns:
            bool: True if synonyms exist
        """
        return self._d_synonyms_fwd and len(self._d_synonyms_fwd)

    def is_canon(self,
                 input_text: str) -> bool:
        """Check if Input Text is Canonical Entity

        Args:
            input_text (str): any input string

        Returns:
            bool: True if the input string is a Nursing Entity
        """
        if not self._has_synonyms_fwd():
            return False

        return self._cleanse_canon(input_text) in self._d_synonyms_fwd

    def is_variant(self,
                   input_text: str) -> bool:
        """Check if Input Text is known variant for at least one Canonical Entry

        Args:
            input_text (str): any input string

        Returns:
            bool: True if the input string is a known synonym to a Nursing Entity
        """
        if not self._has_synonyms_rev():
            return False

        return self._cleanse_variant(input_text) in self._d_synonyms_rev

    def find_canon(self,
                   input_text: str) -> str or None:
        """Find the Canonical Representation of the Input String

        Args:
            input_text (str): any input string

        Returns:
            str or None: The Canonical Entity
        """

        input_text = self._cleanse_canon(input_text)
        if not input_text or not len(input_text):
            return None

        def find() -> str or None:

            # is canon
            if self._has_synonyms_fwd() and input_text in self._d_synonyms_fwd:
                return input_text

            # is variant
            if self._has_synonyms_rev():
                if input_text in self._d_synonyms_rev:
                    return self._d_synonyms_rev[input_text]
                if '_' in input_text:
                    temp = input_text.replace('_', ' ')
                    if temp in self._d_synonyms_rev:
                        return self._d_synonyms_rev[temp]

        result = find()
        if not result or not len(result):
            return None  # no canonical form exists

        def get_result() -> str:
            if type(result) == list:
                if len(result) > 1:
                    self.logger.warning('\n'.join([
                        f'Multi Typed Result (total={len(result)})',
                        f'\tInput: {input_text}',
                        f'\tTypes: {result}']))
                return result[0]
            return result

        result = get_result()
        if self.isEnabledForDebug:
            Enforcer.is_str(result)

        return result

    def find_variants(self,
                      input_text: str) -> list:
        """Find the Synonyms for a known Entity

        Args:
            input_text (str): any input string

        Returns:
            list or None: a list of synonyms for the input entity
        """

        input_text = self._cleanse_variant(input_text)
        if not input_text or not len(input_text):
            return None

        def get_values() -> list:

            # is canon

            if self._has_synonyms_fwd() and input_text in self._d_synonyms_fwd:
                return self._d_synonyms_fwd[input_text]

            # is variant
            if self._has_synonyms_rev():
                if input_text in self._d_synonyms_rev:
                    s = set()
                    for canon in self._d_synonyms_rev[input_text]:
                        [s.add(x) for x in self._d_synonyms_fwd[canon]]
                    return sorted(s, key=len)

            return []

        values = [x for x in get_values() if x and len(x)]

        return values



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
""" Connect to an RDF Graph (Ontology) """


import os

from baseblock import BaseObject, FileIO
from rdflib import Graph, Namespace


class OwlGraphConnector(BaseObject):
    """ Connect to an RDF Graph (Ontology) """

    def __init__(self,
                 prefix: str,
                 namespace: str,
                 ontology_name: str,
                 absolute_path: str):
        """ Change Log

        Created:
            6-Oct-2021
            craigtrim@gmail.com
            *   Create Owl2PY Util Service
        Updated:
            25-May-2022
            craigtrim@gmail.com
            *   refactor into 'ask-owl' repo
                https://github.com/craigtrim/askowl/issues/1

        Args:
            prefix (str): the query prefix
            namespace (str): the ontology namespace
            ontology_name (str): the ontology name (phsyical file name)
            absolute_path (str): the absolute path to the OWL model
        """
        BaseObject.__init__(self, __name__)

        self._format = 'ttl'
        self._prefix = prefix
        self._namespace = namespace
        self._ontology_name = ontology_name
        self._absolute_path = absolute_path

        self._graph = self._process()

        if self.isEnabledForDebug:
            self.logger.debug('\n'.join([
                'Loading Ontology',
                f'\tNamespace: {self._namespace}',
                f'\tName: {self._ontology_name}',
                f'\tPrefix: {self._prefix}',
                f'\tFormat: {self._format}',
                f'\tAbsolute Path: {self._absolute_path}']))

    def _get_versioned_model(self) -> str or None:
        """ Get Versioned Model from Directory

        Notes:
            Ontology Models may be saved as
                /home/path/temp/my_model-0.1.0.owl

            But the only params passed by the consumer are:
                absolute-path:  /home/path/temp
                ontology-name:  my_model.owl

            This function will find any Ontology Model at the given location
                with a matching name, but also with a version

        Returns:
            str or None: the qualified input path to a versioned model, if any
        """
        files = FileIO.load_files(
            directory=self._absolute_path,
            extension='owl')

        if len(files) != 1:
            return None

        basename = os.path.basename(files[0]).split('.')[0].strip()
        if not basename.startswith(self._ontology_name.split('.')[0]):
            return None

        input_path = files[0]
        if self.isEnabledForDebug:
            self.logger.debug('\n'.join([
                'Using Versioned Ontology Model',
                f'\tIncoming Ontology Name: {self._ontology_name}',
                f'\tVersioned Ontology Model: {input_path}']))

        return input_path

    def _process(self) -> Graph:
        """ Load the OWL Model from disk as an RDF Graph

        Returns:
            Graph: an instantiated and in-memory RDF Graph
        """
        g = Graph()

        input_path = os.path.normpath(
            os.path.join(
                self._absolute_path,
                self._ontology_name))

        if not FileIO.exists(input_path):
            input_path = self._get_versioned_model()

        g.parse(input_path,
                format=self._format)

        g.bind(self._prefix,
               Namespace(self._namespace))

        return g

    def graph(self) -> Graph:
        return self._graph



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
""" View Generator: Perform Synonym Transformation """


from collections import defaultdict

from baseblock import BaseObject, TextUtils


class GenerateViewSynonyms(BaseObject):
    """ View Generator: Perform Synonym Transformation """

    __punkt = [
        '!',
        '?',
        '.',
    ]

    def __init__(self):
        """ Change Log
        Created:
            7-Oct-2021
            craigtrim@gmail.com
            *   Create Owl2PY Util Service
        Updated:
            2-Feb-2022
            craigtrim@gmail.com
            *   augment forms by removing punctuation
                Defect in Synonym Swapping when Punctuation is Present
        Updated:
            27-May-2022
            craigtrim@gmail.com
            *   ported to ask-owl
                https://github.com/craigtrim/askowl/issues/6
        Updated:
            2-Jun-2022
            craigtrim@gmail.com
            *   tokenize spaces in synonyms
                https://github.com/craigtrim/askowl/issues/7
        Updated:
            25-Nov-2022
            craigtrim@gmail.com
            *   remove '+' from spanned synonyms
                https://github.com/craigtrim/owl-finder/issues/4#issuecomment-1327975311
        """
        BaseObject.__init__(self, __name__)

    @staticmethod
    def _reverse(d: dict) -> dict:
        d_rev = defaultdict(list)
        for k in d:
            for v in d[k]:
                d_rev[v].append(k)

        return dict(d_rev)

    def process(self,
                d_results: dict,
                reverse: bool = False) -> dict:
        d = {}

        for k in d_results:

            s = set()
            for value in d_results[k]:

                # Represent Spanned Synonyms into String-Based Synonyms
                if '+' in value:  # https://github.com/craigtrim/owl-finder/issues/4#issuecomment-1327975311
                    value = value.replace('+', ' ')

                s.add(value.lower())

                # Reference: Defect in Synonym Swapping when Punctuation is Present
                # https://github.com/craigtrim/askowl/issues/11
                for punkt in self.__punkt:
                    if value.endswith(punkt):
                        s.add(value[:len(value) - len(punkt)])

                # Reference: Tokenization of Space Required
                # https://github.com/craigtrim/askowl/issues/7
                # TODO: likely need better testing around period vs ellipses vs multiple periods here
                if '.' in value and '...' not in value:
                    _value = value.replace('.', ' . ')
                    _value = TextUtils.update_spacing(_value)
                    s.add(_value)

            d[k.lower()] = sorted(s, key=len)

        if reverse:
            return self._reverse(d)

        return d



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['owl_finder',
 'owl_finder.multiquery',
 'owl_finder.multiquery.bp',
 'owl_finder.multiquery.dmo',
 'owl_finder.multiquery.dmo.span',
 'owl_finder.multiquery.svc',
 'owl_finder.singlequery',
 'owl_finder.singlequery.bp',
 'owl_finder.singlequery.dmo',
 'owl_finder.singlequery.dto',
 'owl_finder.singlequery.svc']

package_data = \
{'': ['*']}

install_requires = \
['baseblock', 'rdflib>=6.2.0,<7.0.0']

setup_kwargs = {
    'name': 'owl-finder',
    'version': '0.1.9',
    'description': 'Perform Ontology Queries with a Finder Facade',
    'long_description': '# Ontology Finder (owl-finder)\nA Finder Facade for Querying Ontology Files\n',
    'author': 'Craig Trim',
    'author_email': 'craigtrim@gmail.com',
    'maintainer': 'Craig Trim',
    'maintainer_email': 'craigtrim@gmail.com',
    'url': 'https://github.com/craigtrim/owl-finder',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8.5,<4.0.0',
}


setup(**setup_kwargs)



from .find_ner import FindNER
from .find_synonyms import FindSynonyms
from .find_types import FindTypes
from .load_synonyms import LoadSynonyms
from .query_ner_depth import QueryNerDepth
from .query_ner_label import QueryNerLabel
from .query_ner_taxo import QueryNerTaxo



# !/usr/bin/env python
# -*- coding: UTF-8 -*-
""" Perform Runtime KB Transformation """


from baseblock import BaseObject


class OwlSpanAugment(BaseObject):
    """Augment Lookup with Possible Permuations

    Sample Input:
        {
            2: [
                'clinical outcome',
                'outcome evaluation',
            ],
            3: [
                'clinical outcome evaluation'
            ]
        }

    Sample Output:
        {
            2: [
                'clinical outcome',
                'clinical_outcome evaluation',
                'clinical outcome_evaluation',
                'outcome evaluation',
            ],
            3: [
                'clinical outcome evaluation'
            ]
        }

    Purpose:
        given this input text:
            "clinical outcome evaluations"

        there is no trigram for this exact match

        these swaps occur:
            "clinical outcome"      -->     "clinical_outcome"
            "evaluations"           -->     "evaluation"

        which leaves
            clinical_outcome evaluation

        but this term does not match
            clinical outcome evaluation

        so this algorithm will augment the possible choices by
            adding in possible swaps that can occur in the context
            of longer gram choices
    """

    __blacklist = [
        'and',
        'for',
        'of',
        'to',
    ]

    def __init__(self):
        """ Change Log

        Created:
            15-Oct-2021
            craig.@graffl.ai
            *   https://github.com/grafflr/graffl-core/issues/54
        Updated:
            26-May-2022
            craigtrim@gmail.com
            *   migrated to ask-owl in pursuit of
                https://github.com/grafflr/ask-owl/issues/4
        """
        BaseObject.__init__(self, __name__)

    def _to_result_set(self,
                       buffer: list) -> list:

        def is_valid(a_tuple: tuple) -> bool:
            for item in self.__blacklist:
                if item in a_tuple:
                    return False
            return True

        results = set()

        buffer = [x for x in buffer if is_valid(x)]
        [results.add(' '.join(x).strip()) for x in buffer]

        return sorted(results, key=len)

    def _grams_by_n(self,
                    n: int,
                    values: list) -> list:
        buffer = []

        for value in values:
            tokens = value.split(' ')
            for i in range(len(tokens)):
                if i + n < len(tokens):
                    gramresult = []
                    j = 0
                    while j < n:
                        gramresult.append(tokens[i + j])
                        j += 1
                    buffer.append(gramresult)

        return self._to_result_set(buffer)

    def process(self,
                d_gramsize: dict) -> dict:

        gram2 = set(d_gramsize[2])
        gram3 = set(d_gramsize[3])
        gram4 = set(d_gramsize[4])
        gram5 = set(d_gramsize[5])

        def augment(value: str) -> None:
            size = len(value.split(' '))
            if size == 2:
                gram2.add(value)
            elif size == 3:
                gram3.add(value)
            elif size == 4:
                gram4.add(value)
            elif size == 5:
                gram5.add(value)

        def iterate(func, n, values) -> None:
            grams = func(n, values)
            for value in values:
                for gram in grams:
                    if gram in value:

                        _gram = gram.replace(' ', '_')
                        _value = value.replace(gram, _gram)
                        augment(_value)

        iterate(self._grams_by_n, 5, d_gramsize[6])
        iterate(self._grams_by_n, 4, d_gramsize[5])
        iterate(self._grams_by_n, 3, d_gramsize[4])
        iterate(self._grams_by_n, 2, d_gramsize[3])

        d_gramsize[2] = gram2
        d_gramsize[3] = gram3
        d_gramsize[4] = gram4
        d_gramsize[5] = gram5

        return dict(d_gramsize)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
from enum import Enum


class QueryResultType(Enum):

    DO_NOT_TRANSFORM = 0

    LIST_OF_STRINGS = 10

    DICT_OF_STR2STR = 20
    DICT_OF_STR2LIST = 21
    DICT_OF_STR2DICT = 22
